# Copyright 2019 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Backup Driver for Inspur InStorage
"""

import datetime
import errno
import json
import os
import os.path
import pytz
import re
import six
import stat


from eventlet import tpool
from os_brick.remotefs import remotefs as remotefs_brick
from oslo_concurrency import processutils as putils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import units

from cinder.utils import exception
from cinder.i18n import _
from cinder import interface
from cinder.backup import driver
from cinder import utils


from cinder.backup.drivers.inspur.instorage.cli_helper import InStorageSSH
from cinder.backup.drivers.inspur.instorage.cli_helper import SSHRunner

from cinder import coordination
def synchronized_lock(lockname):
    return coordination.synchronized(lockname)

instorage_opts = [
    cfg.HostAddressOpt(
        'instorage_backup_ip',
        required=True,
        help='IP address for the InStorage.'
    ),
    cfg.PortOpt(
        'instorage_backup_port',
        default=22,
        help='Port number for the InStorage.'
    ),
    cfg.StrOpt(
        'instorage_backup_login',
        required=True,
        help='Username for the InStorage.'
    ),
    cfg.StrOpt(
        'instorage_backup_password',
        required=True,
        secret=True,
        help='Password for the InStorage.'
    ),
    cfg.StrOpt(
        'instorage_backup_pool',
        required=True,
        help='The storage pool where the backup device is created.'
    ),
    cfg.IntOpt(
        'instorage_backup_iogrp',
        help='The IOGrp the backup volume use.'
    ),
    cfg.StrOpt(
        'instorage_backup_link_protocol',
        required=True,
        help='The link protocol used to attach backup volume to host. '
             'iscsi for iSCSI and fibre_channel for FC.'
    ),
    cfg.IntOpt(
        'instorage_backup_initial_size',
        default=2,
        min=2, max=100,
        help='The initial size of the backup device.'
    ),
    cfg.IntOpt(
        'instorage_backup_transmit_chunk_size',
        default=256,
        min=32, max=256,
        help='Transmit chunk size to do differential check and copy, '
             'can only be 32/64/128/256, the unit is KiB.'
    ),
    cfg.BoolOpt(
        'instorage_backup_use_multipath',
        default=False,
        help='Use multipath when connect the backup device to host.'
    ),
    cfg.StrOpt(
        'instorage_backup_meta_mount_base',
        required=True,
        help='Base dir containing mount point for NFS share.'
    ),
    cfg.StrOpt(
        'instorage_backup_meta_share_path',
        help='NFS share in hostname:path, ipv4addr:path, '
             'or "[ipv6addr]:path" format.'
    ),
    cfg.StrOpt(
        'instorage_backup_meta_mount_options',
        help='Mount options passed to the NFS client. '
             'See NFS man page for details.'
    ),
]

CONF = cfg.CONF
CONF.register_opts(instorage_opts)
LOG = log.getLogger(__name__)


class InStorageMetadataBackup(object):

    METADATA_BACKUP_VERSION = "1.0.0"

    def __init__(self, backup_path):
        self.backup_path = backup_path

    def save_metadata(self, backup, volume_meta, extra_metadata=None):
        container = self.get_container_name(backup)
        self.put_container(container)

        filename = self.metadata_filename(backup)
        LOG.debug('save metadata started, container name: %(container)s,'
                  ' metadata filename: %(filename)s.',
                  {'container': container, 'filename': filename})
        metadata = {}
        metadata['version'] = self.METADATA_BACKUP_VERSION
        metadata['backup_id'] = backup['id']
        metadata['volume_id'] = backup['volume_id']
        metadata['backup_name'] = backup['display_name']
        metadata['backup_description'] = backup['display_description']
        metadata['created_at'] = str(backup['created_at'])
        metadata['volume_meta'] = volume_meta
        if extra_metadata:
            metadata['extra_metadata'] = extra_metadata
        metadata_json = json.dumps(metadata, sort_keys=True, indent=2)
        if six.PY3:
            metadata_json = metadata_json.encode('utf-8')
        try:
            with self.get_object_writer(container, filename) as writer:
                writer.write(metadata_json)
        except:
            LOG.error('save metadata failed. Metadata file not found:')
        else:
            LOG.debug('save metadata finished. Metadata: %s.', metadata_json)

    def load_metadata(self, backup):
        container = self.get_container_name(backup)
        filename = self.metadata_filename(backup)
        LOG.debug('load metadata started, container name: %(container)s, '
                  'metadata filename: %(filename)s.',
                  {'container': container, 'filename': filename})
        try:
            with self.get_object_reader(container, filename) as reader:
                metadata_json = reader.read()
            if six.PY3:
                metadata_json = metadata_json.decode('utf-8')
            metadata = json.loads(metadata_json)
            LOG.debug('_read_metadata finished. Metadata: %s.', metadata_json)
            return metadata
        except:
            raise

    def delete_metadata(self, backup):
        container = self.get_container_name(backup)
        filename = self.metadata_filename(backup)
        self.delete_object(container, filename)

    def get_container_name(self, backup):
        backup_id = backup['id']
        return os.path.join(backup_id[0:2], backup_id[2:4])

    def metadata_filename(self, backup):
        filename = 'backup_%s_metadata' % backup['id']
        return filename

    def put_container(self, container):
        path = os.path.join(self.backup_path, container)
        if not os.path.exists(path):
            os.makedirs(path)
            permissions = (
                stat.S_IRUSR |
                stat.S_IWUSR |
                stat.S_IXUSR |
                stat.S_IRGRP |
                stat.S_IWGRP |
                stat.S_IXGRP)
            os.chmod(path, permissions)

    def get_object_writer(self, container, object_name):
        path = os.path.join(self.backup_path, container, object_name)
        f = open(path, 'wb')
        permissions = (
            stat.S_IRUSR |
            stat.S_IWUSR |
            stat.S_IRGRP |
            stat.S_IWGRP)
        os.chmod(path, permissions)
        return tpool.Proxy(f)

    def get_object_reader(self, container, object_name):
        path = os.path.join(self.backup_path, container, object_name)
        f = open(path, 'rb')
        return tpool.Proxy(f)

    def delete_object(self, container, object_name):
        # TODO(tbarron):  clean up the container path if it is empty
        path = os.path.join(self.backup_path, container, object_name)
        try:
            os.remove(path)
        except OSError as e:
            # Ignore exception if path does not exist.
            if e.errno != errno.ENOENT:
                raise

@interface.backupdriver
class InStorageBackupDriver(driver.BackupDriver):
    """Inspur InStorage Backup driver.

    .. code::none
        Version history:
            1.0.0 - Initial driver.
                    Driver support:
                        backup create, delete, restore.
                        support full and incremental backup.
    """

    VENDOR = 'INSPUR'
    VERSION = '1.0.0'


    def __init__(self, context, db=None):
        super(InStorageBackupDriver, self).__init__(context, db)

        self.backup_option = dict(
            backup_link_proto=CONF.instorage_backup_link_protocol,
            backup_pool=CONF.instorage_backup_pool,
            backup_thin_rsize=CONF.instorage_backup_initial_size,
            backup_thin_grainsize=CONF.instorage_backup_transmit_chunk_size,
        )
        if CONF.instorage_backup_iogrp is not None:
            self.backup_option.update(dict(
                backup_iogrp=CONF.instorage_backup_iogrp,
            ))

        # same with the thin volume grain size
        self.transmit_chunk_size = (
            CONF.instorage_backup_transmit_chunk_size * units.Ki
        )
        self.transmit_use_multipath = (
            CONF.instorage_backup_use_multipath
        )

        self.ssh_runner = SSHRunner(**dict(
            host=CONF.instorage_backup_ip,
            port=CONF.instorage_backup_port,
            login=CONF.instorage_backup_login,
            password=self._decode_pwd(CONF.instorage_backup_password)
        ))

        self.assistant = InStorageAssistant(
            self.ssh_runner, self.backup_option
        )

        self.meta_backup_path = self._init_backup_meta_share(
            CONF.instorage_backup_meta_mount_base,
            CONF.instorage_backup_meta_share_path,
            CONF.instorage_backup_meta_mount_options
        )

    def _decode_pwd(self,encode_pwd):
        cmd = "echo %s | base64 -d" % encode_pwd  
        handler = os.popen(cmd)
        decoded_pwd = handler.readline()
        return decoded_pwd.replace('\n','')

    def check_for_setup_error(self):
        # check the pool
        pools = self.assistant.get_available_pools()
        if self.backup_option['backup_pool'] not in pools:
            msg = (_('Pool %s not exist on the storage system.') %
                   self.backup_option['backup_pool'])
            raise exception.InvalidParameterValue(msg)

        # check the grain chunk size
        valid_chunk_size = set([32, 64, 128, 256])
        if self.backup_option['backup_thin_grainsize'] not in valid_chunk_size:
            msg = (_('Transmit chunk size can only be 32/64/128/256'))
            raise exception.InvalidParameterValue(msg)

        # check the link protocol
        link_protos = set(['iscsi', 'fibre_channel'])
        if self.backup_option['backup_link_proto'] not in link_protos:
            raise exception.InvalidParameterValue(
                _('Backup link protocol only support %s'), link_protos
            )

        if 'backup_iogrp' in self.backup_option:
            iogrps = self.assistant.get_available_iogrps()
            if self.backup_option['backup_iogrp'] not in iogrps:
                raise exception.InvalidParameterValue(
                    _('IOgrp can only be one of %s'), iogrps
                )

        # check the meta backup path
        if self.meta_backup_path is None:
            raise exception.InvalidParameterValue(
                _('Share cannot be mount.')
            )

    def _init_backup_meta_share(self, mount_base, share_path, mount_options):
        if (mount_base is None or mount_base == ''
            or not os.path.exists(mount_base)):
            return

        mount_path = mount_base

        root_helper = utils.get_root_helper()
        if share_path:
            remotefsclient = remotefs_brick.RemoteFsClient(
                'nfs',
                root_helper,
                nfs_mount_point_base=mount_base,
                nfs_mount_options=mount_options
            )
            remotefsclient.mount(share_path)

            mount_path = remotefsclient.get_mount_point(share_path)

        # Ensure we can write to this share
        group_id = os.getegid()
        current_group_id = utils.get_file_gid(mount_path)
        current_mode = utils.get_file_mode(mount_path)

        if group_id != current_group_id:
            cmd = ['chgrp', '-R', group_id, mount_path]
            putils.execute(*cmd, root_helper=root_helper, run_as_root=True)

        if not (current_mode & stat.S_IWGRP):
            cmd = ['chmod', '-R', 'g+w', mount_path]
            putils.execute(*cmd, root_helper=root_helper, run_as_root=True)

        return mount_path

    def _attach_device(self, volume):
        
        properties = utils.brick_get_connector_properties(
            multipath=self.transmit_use_multipath
        )
        
        conn = self.assistant.initialize_connection(volume, properties)

        connector = utils.brick_get_connector(
            conn['driver_volume_type'],
            use_multipath=properties.get('multipath', False),
            device_scan_attempts=5,
            conn=conn,
            expect_raw_disk=True
        )
        vol_handle = connector.connect_volume(conn['data'])

        attach_info = {
            'conn': conn,
            'device': vol_handle,
            'connector': connector,
            'properties': properties
        }
        return attach_info

    def _detach_device(self, volume, attach_info):
        connector = attach_info['connector']
        connector.disconnect_volume(
            attach_info['conn']['data'],
            attach_info['device'],
        )
        self.assistant.terminate_connection(volume, attach_info['properties'])
    
    def _transmit_data(self, src_file, dst_file, diff=False):
        total_read = 0
        total_write = 0
        time_start = datetime.datetime.now()

        while True:
            data_src = src_file.read(self.transmit_chunk_size)
            if data_src == b'':
                # game over
                break
            total_read += 1

            if diff:
                data_dst = dst_file.read(self.transmit_chunk_size)
                if data_dst == data_src:
                    continue

            # go back and do a write to the dst file
            dst_file.seek(src_file.tell() - len(data_src))
            dst_file.write(data_src)
            total_write += 1

        time_end = datetime.datetime.now()
        delta = time_end - time_start
        LOG.debug('%s transmit with read %d and write %d in %d seconds.',
                  'Diff' if diff else 'Full', total_read, total_write,
                   delta.total_seconds())

    def _get_proper_base_volume(self, backup):
        """Get the most recent backup base volume.

        get all the backup chains the for the volume to backup,
        and compare the head of the base volumes, use the one
        closest to the backup, but not exceed the backup.

        base volume is a dict like follows
        {
            'id': 'the source volume id of the backup',
            'size': 'the source volume size of the backup',
            'name': 'the base volume name'
        }
        """
        base_volume = {
            'type': 'base',
            'id': backup.volume_id,
            'size': backup.size,
            'name': None,
            'timestamp': backup.data_timestamp
        }
        relations = self.assistant.obtain_backup_relation_map(base_volume)
        min_diff = datetime.timedelta.max
        min_parent = None
        min_base = None
        for k, v in relations.items():
            diff = datetime.timedelta.max
            obj = v[0]
            if backup.data_timestamp > obj['timestamp']:
                diff = backup.data_timestamp - obj['timestamp']

            if diff < min_diff:
                min_diff = diff
                min_parent = obj
                min_base = k

        if min_parent != None:
            base_volume['name'] = min_base
            # compare the size of base volume and the size of origin volume,if not equal,then create a new base volume
            origin_volume_size = backup.size
            attr = self.assistant.get_vdisk_attributes(base_volume['name'])
            self.assistant.get_vdisk_attributes
            base_volume_size = attr['capacity']
            if(int(origin_volume_size) * 1024 * 1024 * 1024 == int(base_volume_size)):
                return base_volume

        # not found an exist base volume, we create a new base volume
        base_volume['name'] = self.assistant.generate_base_vol_name(
            base_volume, with_timestamp=True
        )
        return self.assistant.create_base_volume(base_volume)

    def _get_backup_volume(self, backup):
        backup_volume = {
            'type': 'backup',
            'id': backup.id,
            'name': None,
            'size': backup.size,
            'timestamp': backup.data_timestamp
        }
        backup_volume['name'] = self.assistant.generate_backup_vol_name(backup_volume)

        return backup_volume

    def _backup_metadata(self, backup):
        volume_meta = self.get_metadata(backup['volume_id'])
        if not volume_meta:
            LOG.debug('No volume metadata to backup.')
            return

        meta_backup = InStorageMetadataBackup(self.meta_backup_path)
        meta_backup.save_metadata(backup, volume_meta)

    @utils.trace
    def backup(self, backup, volume_file, backup_metadata=True):
        #1. find the most recent backup base volume.
        #2. if not get one, create a new one.
        base_volume = self._get_proper_base_volume(backup)
        backup_volume = self._get_backup_volume(backup)

        #3. attach the base volume.
        attach_info = self._attach_device(base_volume)

        #4. do a differential write.
        backup_device_path = attach_info['device']['path']
        try:
            with utils.temporary_chown(backup_device_path):
                with open(backup_device_path, 'rb+') as backup_file:
                    self._transmit_data(volume_file, tpool.Proxy(backup_file), True)
            LOG.debug('Finish backup data transmit from '
                    'backup source to backup device.')
        except:
            LOG.debug(' failed to backup data transmit from '
                    'backup source to backup device.')
            raise

        #5. detach the base volume
        self._detach_device(base_volume, attach_info)

        #6. create a snapshot as the backup
        self.assistant.create_backup(base_volume, backup_volume)

        #7. save backup metadata
        if not backup_metadata:
            return

        try:
            self._backup_metadata(backup)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Backup volume metadata failed.")
                self.delete_backup(backup)
            raise

    def _load_metadata(self, backup):
        meta_backup = InStorageMetadataBackup(self.meta_backup_path)
        metadata = meta_backup.load_metadata(backup)
        volume_meta = metadata.get('volume_meta', None)
        return volume_meta

    @utils.trace
    def restore(self, backup, volume_id, volume_file):
        #1. find the backup volume.
        backup_volume = self._get_backup_volume(backup)

        #2. attach the volume.
        attach_info = self._attach_device(backup_volume)

        #3. do a full transmit.
        backup_device_path = attach_info['device']['path']
        try:
            with utils.temporary_chown(backup_device_path):
                with open(backup_device_path, 'rb') as backup_file:
                    self._transmit_data(tpool.Proxy(backup_file), volume_file)
            LOG.debug('Finish restore data transmit from %s to %s.' % (
                backup_volume['name'], volume_id
            ))
        except:
            LOG.error('failed to restore data transmit from %s to %s.' % (
                backup_volume['name'], volume_id
            ))
            raise

        #4. detach the volume.
        self._detach_device(backup_volume, attach_info)

        #5. restore the volume meta
        try:
            volume_meta = self._load_metadata(backup)
            if not volume_meta:
                LOG.debug('No volume metadata in this backup.')
                return
            self.put_metadata(volume_id, volume_meta)
        except exception.BackupMetadataUnsupportedVersion:
            msg = _("Metadata restore failed due to incompatible version.")
            LOG.error(msg)
            raise exception.BackupOperationError(msg)

    def _delete_metadata(self, backup):
        meta_backup = InStorageMetadataBackup(self.meta_backup_path)
        meta_backup.delete_metadata(backup)

    def delete(self, backup):
        """Delete is the method to delete a backup in old openstack."""
        return self.delete_backup(backup)

    @utils.trace
    def delete_backup(self, backup):
        #1. find the backup volume.
        backup_volume = self._get_backup_volume(backup)
        attr = self.assistant.get_vdisk_attributes(backup_volume['name'])
        if attr is None:
            LOG.info('backup volume %s not exist, just return.', backup_volume['name'])
            return

        #2. find the base volume.
        relations = self.assistant.obtain_backup_relation_map(backup_volume=backup_volume)

        #3. delete the backup volume.
        self.assistant.delete_backup(backup_volume, relations)

        #4. if no backup volume depend the base volume, delete it.
        # although follows is a loop, there is only one key value pair.
        for k, v in relations.items():
            if len(v) == 1:
                # as we have delete the only lc relationship when delete backup,
                # now we can delete the base volume
                base_volume = {
                    'name': k
                }
                LOG.debug('Delete the base volume %s.', k)
                self.assistant.delete_base_volume(base_volume)

        #5. delete the meta backup
        self._delete_metadata(backup)

def get_backup_driver(context):
    return InStorageBackupDriver(context)

class InStorageAssistant(object):

    BASE_BACKUP_VOLUME_NAME_PATTERN = 'backup.base.%(volume_id)s.%(timestamp)s'
    BACKUP_VOLUME_NAME_PATTERN = 'backup.%(backup_id)s.%(timestamp)s'
    TIMESTAMP_PATTERN = '%Y%m%d%H%M%S'

    def __init__(self, ssh_runner, backup_option):
        self.ssh = InStorageSSH(ssh_runner)
        self.backup_volume_option = backup_option
        # link protocal can be 'fibre_channel' or 'iscsi'
        self.link_protocal = backup_option['backup_link_proto']

    @staticmethod
    def handle_keyerror(cmd, out):
        msg = (_('Could not find key in output of command %(cmd)s: %(out)s.')
               % {'out': out, 'cmd': cmd})
        raise exception.ShareBackendException(msg=msg)

    def generate_vdisk_create_parameter(self):
        params = []
        # set the pool
        params.extend(
            ['-mdiskgrp', self.backup_volume_option['backup_pool']]
        )

        # set iogrp if setted
        if 'backup_iogrp' in self.backup_volume_option:
            params.extend(
                ['-iogrp', '%d' % self.backup_volume_option['backup_iogrp']]
            )

        # set the real size of a thin volume
        params.extend([
            '-rsize',
            '%d%%' % self.backup_volume_option['backup_thin_rsize']
        ])
        # set the grain size
        params.extend([
            '-grainsize',
            '%d' % self.backup_volume_option['backup_thin_grainsize']
        ])
        params.extend(['-autoexpand'])

        return params

    def size_to_gb(self, size):
        new_size = 0

        if 'P' in size:
            new_size = int(float(size.rstrip('PB')) * units.Mi)
        elif 'T' in size:
            new_size = int(float(size.rstrip('TB')) * units.Ki)
        elif 'G' in size:
            new_size = int(float(size.rstrip('GB')) * 1)
        elif 'M' in size:
            mb_size = float(size.rstrip('MB'))
            new_size = int((mb_size + units.Ki - 1) / units.Ki)

        return new_size

    def get_available_pools(self):
        pools = self.ssh.lsmdiskgrp()
        return [p['name'] for p in pools]

    def get_available_iogrps(self):
        iogrps = self.ssh.lsiogrp()
        valid_iogrp = set()
        for iogrp in iogrps:
            if int(iogrp['node_count']) > 0 and int(iogrp['vdisk_count']) > 0:
                valid_iogrp.add(int(iogrp['id']))

        return list(valid_iogrp)

    @staticmethod
    def fixup_datetime(timestamp):
        LOG.debug('time to fixup %s.', timestamp)
        microsecond = timestamp.microsecond
        timestamp = timestamp.replace(microsecond=0)
        if microsecond >= 500000:
            timestamp = timestamp + datetime.timedelta(0, 1, 0)

        LOG.debug('time fixup as %s.', timestamp)
        return timestamp

    def generate_base_vol_name(self, base_volume, with_timestamp=False):
        data = {
            'volume_id': base_volume['id'],
            'timestamp': ''
        }
        if with_timestamp:
            new_time = self.fixup_datetime(base_volume['timestamp'])
            data['timestamp'] = new_time.strftime(self.TIMESTAMP_PATTERN)

        return  self.BASE_BACKUP_VOLUME_NAME_PATTERN % data

    def generate_backup_vol_name(self, backup_volume):
        new_time = self.fixup_datetime(backup_volume['timestamp'])
        data = {
            'backup_id': backup_volume['id'],
            'timestamp': new_time.strftime(self.TIMESTAMP_PATTERN)
        }
        return  self.BACKUP_VOLUME_NAME_PATTERN % data

    @utils.trace
    def obtain_backup_relation_map(self, base_volume=None, backup_volume=None):
        """List and arrange the local copy relation maps of a backup suits.

        Get them either by a backup or a backup's base volume infor.
        {
            "base-volume-1": [{"lcname":"lc01", "name": "backup-volume-t0"}],
            "base-volume-2": [{"lcname":"lc02", "name": "backup-volume-t1"}]
        }

        if base volume is given, will get all the map info.
        if only backup is given, will only get the chain in which backup is in.
        the backups of the list is in descend order by backup time. 

        base volume is a dict like {"id": "xxxxx"}
        backup volume is a dict like {"id": "xxxxx"}
        """
        if base_volume:
            prefix = self.generate_base_vol_name(base_volume)
            prefix = prefix + "*"
        elif backup_volume:
            name = self.generate_backup_vol_name(backup_volume)
            lcmaps = self.ssh.lslcmap('target_vdisk_name', name)
            prefix = lcmaps[0]['source_vdisk_name']
        else:
            raise exception.InvalidParameterValue()

        lcmaps = self.ssh.lslcmap('source_vdisk_name', prefix)

        relations = dict()
        for lcmap in lcmaps:
            base_name = lcmap['source_vdisk_name']
            backup_name = lcmap['target_vdisk_name']
            if base_name not in relations:
                relations[base_name] = []

            timestamp = datetime.datetime.strptime(
                backup_name[-14:], self.TIMESTAMP_PATTERN
            )
            timestamp = timestamp.replace(tzinfo=pytz.timezone('UTC'))
            backup_info = {
                'lcmapid': lcmap['id'],
                'lcmapname': lcmap['name'],
                'vdiskname': backup_name,
                'timestamp': timestamp
            }
            relations[base_name].append(backup_info)

        for k, v in relations.items():
            v.sort(key=lambda x: x['timestamp'], reverse=True)

        LOG.debug('Obtain relations %s.', relations)
        return relations

    def create_backup(self, base_volume, backup_volume):
        """Create a backup on the base volume.

        Now the base volume's contant is same with the backup's contant,
        first create a backup volume,
        then add a LC relationship between them with copyrate = 0,
        which is actually a snapshot of the base volume.
        """
        params = self.generate_vdisk_create_parameter()
        self.ssh.mkvdisk(backup_volume['name'], backup_volume['size'], params)

        # add the local copy relationship with copyrate = 0
        lcmapid = self.ssh.mklcmap(base_volume['name'], backup_volume['name'])
        self.ssh.startlcmap(lcmapid)

        return backup_volume

    @utils.retry(exception.VolumeDriverException, interval=5, retries=36)
    def waiting_lcmap_stopped(self, lcmapid):
        lcmaps = self.ssh.lslcmap('id', lcmapid)
        if lcmaps[0]['status'] == 'stopped':
            return
        else:
            msg = (_('lcmap %s is not stopped.') % lcmapid)
            raise exception.VolumeDriverException(msg)

    def delete_backup(self, backup_volume, relations):
        """Delete the backup.

        first delete the LC relationship between backup and it's base volume,
        then delete the backup volume.
        """
        to_del = None
        for k, v in relations.items():
            for backup_info in v:
                if backup_info['vdiskname'] == backup_volume['name']:
                    to_del = backup_info
                    break

        if to_del is None:
            return

        self.ssh.stoplcmap(to_del['lcmapid'])
        # waiting lcmap status to stopped
        self.waiting_lcmap_stopped(to_del['lcmapid'])
        # and then delete the lcmap
        self.ssh.rmlcmap(to_del['lcmapid'])
        # and then remove the vdisk
        self.ssh.rmvdisk(to_del['vdiskname'])

    def create_base_volume(self, base_volume):
        """Create a base volume for the backups."""
        params = self.generate_vdisk_create_parameter()
        self.ssh.mkvdisk(base_volume['name'], base_volume['size'], params)
        return base_volume
    
    def delete_base_volume(self, base_volume):
        """Delete the base volume."""
        self.ssh.rmvdisk(base_volume['name'])

    def initialize_connection(self, volume, connector):
        """Perform necessary work to make a connection."""
        @synchronized_lock('instorage-host' + connector['host'])
        def _do_initialize_connection_locked():
            return self._do_initialize_connection(volume, connector)
        return _do_initialize_connection_locked()

    def _do_initialize_connection(self, volume, connector):
        connector = connector.copy()

        if self.link_protocal == 'fibre_channel':
            connector.pop('initiator', None)
            return self.initialize_fc_connection(volume, connector)
        elif self.link_protocal == 'iscsi':
            connector.pop('wwpns', None)
            return self.initialize_iscsi_connection(volume, connector)

    def terminate_connection(self, volume, connector):
        """Cleanup after an  connection has been terminated."""

        @synchronized_lock('instorage-host'  + connector['host'])
        def _do_terminate_connection_locked():
            return self._do_terminate_connection(volume, connector)
        return _do_terminate_connection_locked()

    def _do_terminate_connection(self, volume, connector):
        volume_name = volume['name']
        connector = connector.copy()
        if self.link_protocal == 'fibre_channel':
            connector.pop('initiator', None) 
        elif self.link_protocal == 'iscsi':
            connector.pop('wwpns', None)

        info = {
            'driver_volume_type': self.link_protocal,
            'data': {}
        }
        host_name = None
        if 'host' in connector:
            host_name = self.get_host_from_connector(connector)
            if host_name is None:
                msg = (_('terminate_connection: Failed to get host name from'
                         ' connector.'))
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)

        # Unmap volumes, if hostname is None, need to get value from vdiskmap
        host_name = self.unmap_vol_from_host(volume_name, host_name)

        # Host_name could be none
        if host_name:
            # If no volume mapped to the host, we will delete the host.
            vdiskmaps = self.ssh.lshostvdiskmap(host_name)
            if not len(vdiskmaps):
                self.ssh.rmhost(host_name)

        return info

    def initialize_fc_connection(self, volume, connector):
        volume_name = volume['name']

        host_name = self.get_host_from_connector(connector)
        if host_name is None:
            host_name = self.create_host(connector)

        lun_id = self.map_vol_to_host(volume_name, host_name, True)

        properties = {
            'target_discovered': False,
            'target_lun': lun_id
        }

        try:
            conn_wwpns = self.get_conn_fc_wwpns(host_name)

            # If conn_wwpns is empty, then that means that there were
            # no target ports with visibility to any of the initiators
            # so we return all target ports.
            if not conn_wwpns:
                storage_status = self.get_storage_status()
                for k, node_status in storage_status['node_status'].items():
                    conn_wwpns.extend(node_status['WWPN'])

            i_t_map = utils.make_initiator_target_all2all_map(
                connector['wwpns'], conn_wwpns
            )

            properties['target_wwn'] = conn_wwpns
            properties['initiator_target_map'] = i_t_map
        except Exception:
            with excutils.save_and_reraise_exception():
                self.terminate_connection(volume, connector)
                LOG.error('initialize_connection: Failed '
                          'to collect return '
                          'properties for volume %(vol)s and connector '
                          '%(conn)s.\n', {'vol': volume,
                                          'conn': connector})
            raise

        info = {
            'driver_volume_type': self.link_protocal,
            'data': properties
        }
        # fczm_utils.add_fc_zone(info) # do not deal with fc zone
        return info
    
    def initialize_iscsi_connection(self, volume, connector):
        """Perform necessary work to make an iSCSI connection."""
        volume_name = volume['name']
        use_multipath = connector.get('multipath', False)

        host_name = self.get_host_from_connector(connector)
        if host_name is None:
            host_name = self.create_host(connector)

        chap_secret = self.get_chap_secret_for_host(host_name)

        lun_id = self.map_vol_to_host(volume_name, host_name, True)
        try:
            storage_status = self.get_storage_status()
            preferred_node = self.get_vdisk_prefer_node(volume_name, storage_status)
            if preferred_node['ipv4']:
                ipaddr = preferred_node['ipv4'][0]
            else:
                # ipv6 need surround with brackets when it use port
                ipaddr = '[%s]' % preferred_node['ipv6'][0]

            properties = {
                'target_discovered': False,
                'target_iqn': preferred_node['iscsi_name'],
                'target_lun': lun_id,
                'target_portal': '%s:%s' % (ipaddr, '3260'),
            }

            if use_multipath:
                properties['target_iqns'] = []
                properties['target_portals'] = []
                properties['target_luns'] = []
                portals = self.get_iscsi_portals(lun_id, storage_status)
                for iqn, portal, lun in portals:
                    properties['target_iqns'].append(iqn)
                    properties['target_portals'].append(portal)
                    properties['target_luns'].append(lun)
        except Exception:
            with excutils.save_and_reraise_exception():
                self.terminate_connection(volume, connector)
                LOG.error('initialize_connection: Failed '
                          'to collect return '
                          'properties for volume %(vol)s and connector '
                          '%(conn)s.\n', {'vol': volume, 'conn': connector})
            raise
        
        if chap_secret:
            properties.update(auth_method='CHAP',
                              auth_username=connector['initiator'],
                              auth_password=chap_secret,
                              discovery_auth_method='CHAP',
                              discovery_auth_username=connector['initiator'],
                              discovery_auth_password=chap_secret)

        return {'driver_volume_type': 'iscsi', 'data': properties}

    def get_vdisk_prefer_node(self, vdisk_name, storage_status):
        vdisk_attr = self.get_vdisk_attributes(vdisk_name)
        if vdisk_attr is None:
            msg = (_('get_vdisk_prefer_node: Failed to get attributes'
                     ' for vdisk %s.') % vdisk_name)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        try:
            preferred_node_id = vdisk_attr['preferred_node_id']
            ioGrp_id = vdisk_attr['IO_group_id']
        except KeyError as e:
            msg = (_('_get_single_iscsi_data: Did not find expected column'
                     ' name in %(volume)s: %(key)s  %(error)s.'),
                   {'volume': vdisk_name, 'key': e.args[0],
                    'error': e})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Get preferred node and other nodes in I/O group
        preferred_node = None
        io_group_nodes = []
        for k, node in storage_status['node_status'].items():
            if self.link_protocal not in node['enabled_protocols']:
                continue
            if node['IO_group'] != ioGrp_id:
                continue
            io_group_nodes.append(node)
            if node['id'] == preferred_node_id:
                preferred_node = node

        if not len(io_group_nodes):
            msg = (_('_get_single_iscsi_data: No node found in '
                     'I/O group %(gid)s for volume %(vol)s.') % {
                'gid': ioGrp_id, 'vol': vdisk_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not preferred_node:
            # Get 1st node in I/O group
            preferred_node = io_group_nodes[0]
            LOG.warning('_get_single_iscsi_data: Did not find a '
                        'preferred node for volume %s.', vdisk_name)

        return preferred_node

    def get_iscsi_portals(self, lun_id, storage_status):
        iscsi_tupples = []
        for k, node in storage_status['node_status'].items():
            iqn = node['iscsi_name']
            if node['ipv4']:
                portals = ['%s:%s' % (ip, '3260') for ip in node['ipv4']]
            elif node['ipv6']:
                portals = ['[%s]:%s' % (ip, '3260') for ip in node['ipv6']]

            for p in portals:
                iscsi_tupples.append((iqn, p, lun_id))

        return iscsi_tupples

    def get_storage_nodes_info(self):
        """Return dictionary containing information on system's nodes."""
        nodes = {}
        resp = self.ssh.lsnode()
        for node_data in resp:
            try:
                if node_data['status'] != 'online':
                    continue
                node = {}
                node['id'] = node_data['id']
                node['name'] = node_data['name']
                node['IO_group'] = node_data['IO_group_id']
                node['iscsi_name'] = node_data['iscsi_name']
                node['WWNN'] = node_data['WWNN']
                node['status'] = node_data['status']
                node['WWPN'] = []
                node['ipv4'] = []
                node['ipv6'] = []
                node['enabled_protocols'] = []
                nodes[node['id']] = node
            except KeyError:
                self.handle_keyerror('lsnode', node_data)
        return nodes

    def add_iscsi_ip_addrs(self, storage_nodes):
        """Add iSCSI IP addresses to system node information."""
        ports_info = self.ssh.lsportip()
        for port_info in ports_info:
            try:
                link_status = port_info['link_state']
                port_status = port_info['state']
                if port_info['node_id'] not in storage_nodes:
                    continue
                if ((port_status == 'configured' and link_status == 'active')
                     or port_status == 'online'):
                    node = storage_nodes[port_info['node_id']]
                    if len(port_info['IP_address']):
                        node['ipv4'].append(port_info['IP_address'])
                    if len(port_info['IP_address_6']):
                        node['ipv6'].append(port_info['IP_address_6'])
            except KeyError:
                self.handle_keyerror('lsportip', port_info)

    def add_fc_wwpns(self, storage_nodes):
        """Add FC WWPNs to system node information."""
        for key in storage_nodes:
            node = storage_nodes[key]
            wwpns = set(node['WWPN'])
            resp = self.ssh.lsportfc(node_id=node['id'])
            for port_info in resp:
                if (port_info['type'] == 'fc'
                    and port_info['status'] == 'active'):
                    wwpns.add(port_info['WWPN'])
            node['WWPN'] = list(wwpns)
            LOG.info('WWPN on node %(node)s: %(wwpn)s.',
                     {'node': node['id'], 'wwpn': node['WWPN']})

    def get_storage_status(self):
        all_storage_info = self.get_storage_nodes_info()
        self.add_iscsi_ip_addrs(all_storage_info)
        self.add_fc_wwpns(all_storage_info)

        status = {
            'enabled_protocols': set(),
            'node_status': {}
        }
        for k, node in all_storage_info.items():
            if ((len(node['ipv4']) or len(node['ipv6'])) and
                    len(node['iscsi_name'])):
                node['enabled_protocols'].append('iscsi')
                status['enabled_protocols'].add('iscsi')
            if len(node['WWPN']):
                node['enabled_protocols'].append('fibre_channel')
                status['enabled_protocols'].add('fibre_channel')
            if node['enabled_protocols']:
                status['node_status'][k] = node
        
        return status

    @utils.trace
    def get_host_from_connector(self, connector, volume_name=None):
        """Return the InStorage host described by the connector."""
        # If we have FC information, we have a faster lookup option
        host_name = None
        if 'wwpns' in connector:
            for wwpn in connector['wwpns']:
                resp = self.ssh.lsfabric(wwpn=wwpn)
                for wwpn_info in resp:
                    try:
                        if (wwpn_info['remote_wwpn'] and
                            wwpn_info['name'] and
                            wwpn_info['remote_wwpn'].lower() ==
                            wwpn.lower()):
                            host_name = wwpn_info['name']
                            return host_name
                    except KeyError:
                        self.handle_keyerror('lsfabric', wwpn_info)

        def update_host_list(host, host_list):
            idx = host_list.index(host)
            del host_list[idx]
            host_list.insert(0, host)

        # That didn't work, so try exhaustive search
        hosts_info = self.ssh.lshost()
        host_list = list(hosts_info.select('name'))
        # If we have a "real" connector, we might be able to find the
        # host entry with fewer queries if we move the host entries
        # that contain the connector's host property value to the front
        # of the list
        if 'host' in connector:
            # order host_list such that the host entries that
            # contain the connector's host name are at the
            # beginning of the list
            for host in host_list:
                if re.search(connector['host'], host):
                    update_host_list(host, host_list)
        # If we have a volume name we have a potential fast path
        # for finding the matching host for that volume.
        # Add the host_names that have mappings for our volume to the
        # head of the list of host names to search them first
        if volume_name:
            hosts_map_info = self.ssh.lsvdiskhostmap(volume_name)
            hosts_map_info_list = list(hosts_map_info.select('host_name'))
            # remove the fast path host names from the end of the list
            # and move to the front so they are only searched for once.
            for host in hosts_map_info_list:
                update_host_list(host, host_list)
        found = False
        for name in host_list:
            try:
                resp = self.ssh.lshost(host=name)
            except exception.VolumeBackendAPIException as ex:
                LOG.debug("Exception message: %s", ex.msg)
                if 'CMMVC5754E' in ex.msg:
                    LOG.debug("CMMVC5754E found in CLI exception.")
                    # CMMVC5754E: The specified object does not exist
                    # The host has been deleted while walking the list.
                    # This is a result of a host change on the MCS that
                    # is out of band to this request.
                    continue
                # unexpected error so reraise it
                with excutils.save_and_reraise_exception():
                    pass
            if 'initiator' in connector:
                for iscsi in resp.select('iscsi_name'):
                    if iscsi == connector['initiator']:
                        host_name = name
                        found = True
                        break
            elif 'wwpns' in connector and len(connector['wwpns']):
                connector_wwpns = [str(x).lower() for x in connector['wwpns']]
                for wwpn in resp.select('WWPN'):
                    if wwpn and wwpn.lower() in connector_wwpns:
                        host_name = name
                        found = True
                        break
            if found:
                break

        return host_name

    @utils.trace
    def create_host(self, connector):
        """Create a new host on the storage system.

        We create a host name and associate it with the given connection
        information.  The host name will be a cleaned up version of the given
        host name (at most 55 characters), plus a random 8-character suffix to
        avoid collisions. The total length should be at most 63 characters.
        """
        # Before we start, make sure host name is a string and that we have
        # one port at least .
        host_name = connector['host']
        if not isinstance(host_name, six.string_types):
            msg = _('create_host: Host name is not unicode or string.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        ports = []
        if 'initiator' in connector:
            ports.append(['initiator', '%s' % connector['initiator']])
        if 'wwpns' in connector:
            for wwpn in connector['wwpns']:
                ports.append(['wwpn', '%s' % wwpn])
        if not len(ports):
            msg = _('create_host: No initiators or wwpns supplied.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

        # Build a host name for the InStorage host - first clean up the name
        if isinstance(host_name, six.text_type):
            host_name = unicodedata.normalize('NFKD', host_name).encode(
                'ascii', 'replace').decode('ascii')

        for num in range(0, 128):
            ch = str(chr(num))
            if not ch.isalnum() and ch not in [' ', '.', '-', '_']:
                host_name = host_name.replace(ch, '-')

        # InStorage doesn't expect hostname that doesn't starts with letter or
        # _.
        if not re.match('^[A-Za-z]', host_name):
            host_name = '_' + host_name

        # Add a random 8-character suffix to avoid collisions
        rand_id =(''.join(map(lambda xx:(hex(ord(xx))[2:]),os.urandom(8))))[0:16]
        host_name = '%s-%s' % (host_name[:55], rand_id)

        # Create a host with one port
        port = ports.pop(0)
        self.ssh.mkhost(host_name, port[0], port[1])

        # Add any additional ports to the host
        for port in ports:
            self.ssh.addhostport(host_name, port[0], port[1])

        return host_name

    def get_vdisk_attributes(self, vdisk):
        attrs = self.ssh.lsvdisk(vdisk)
        return attrs

    def get_lun_id(self, vdisk_name, host_name):
        maps = self.ssh.lsvdiskhostmap(vdisk_name)

        lun_ids = [m['SCSI_id'] for m in maps if m['host_name'] == host_name]
        return lun_ids[0] if lun_ids else None

    @utils.trace
    def map_vol_to_host(self, volume_name, host_name, multihostmap):
        """Create a mapping between a volume to a host."""
        lun_id = self.get_lun_id(volume_name, host_name)
        if lun_id is None:
            lun_id = self.ssh.mkvdiskhostmap(host_name,
                                             volume_name,
                                             None,
                                             multihostmap)
        return int(lun_id)

    @utils.trace
    def unmap_vol_from_host(self, volume_name, host_name):
        """Unmap the volume and delete the host if it has no more mappings."""
        # Check if the mapping exists
        resp = self.ssh.lsvdiskhostmap(volume_name)
        if not len(resp):
            LOG.warning('unmap_vol_from_host: No mapping of volume '
                        '%(vol_name)s to any host found.',
                        {'vol_name': volume_name})
            return host_name
        if host_name is None:
            if len(resp) > 1:
                LOG.warning('unmap_vol_from_host: Multiple mappings of '
                            'volume %(vol_name)s found, no host '
                            'specified.', {'vol_name': volume_name})
                return
            else:
                host_name = resp[0]['host_name']
        else:
            found = False
            for h in resp.select('host_name'):
                if h == host_name:
                    found = True
            if not found:
                LOG.warning('unmap_vol_from_host: No mapping of volume '
                            '%(vol_name)s to host %(host)s found.',
                            {'vol_name': volume_name, 'host': host_name})
                return host_name
        # We now know that the mapping exists
        self.ssh.rmvdiskhostmap(host_name, volume_name)
        return host_name

    def delete_host(self, host_name):
        self.ssh.rmhost(host_name)

    def check_host_mapped_vols(self, host_name):
        return self.ssh.lshostvdiskmap(host_name)

    def get_conn_fc_wwpns(self, host):
        wwpns = set()
        resp = self.ssh.lsfabric(host=host)
        for wwpn in resp.select('local_wwpn'):
            if wwpn is not None:
                wwpns.add(wwpn)
        return list(wwpns)

    def get_chap_secret_for_host(self, host_name):
        """Generate and store a randomly-generated CHAP secret for the host."""
        resp = self.ssh.lsiscsiauth()
        host_found = False
        for host_data in resp:
            try:
                if host_data['name'] == host_name:
                    host_found = True
                    if host_data['iscsi_auth_method'] == 'chap':
                        return host_data['iscsi_chap_secret']
            except KeyError:
                self.handle_keyerror('lsiscsiauth', host_data)
        if not host_found:
            msg = _('Failed to find host %s.') % host_name
            raise exception.VolumeBackendAPIException(data=msg)
        return None
